// package io.faas;

//import java.io.IOException;
//import java.io.InputStreamReader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
//import java.net.InetSocketAddress;
//import java.net.ServerSocket;
//import java.net.Socket;
//import java.util.Collection;
//import java.util.HashMap;
//import java.util.Map;
//import java.util.Scanner;
import javax.json.*;

import java.io.*;
import java.util.*;
import java.net.*;
import java.util.jar.*;


public class FaasIxServer extends Thread {

  private static final int THREAD_TYPE_MAIN = 0;
  private static final int THREAD_TYPE_RELAY = 1;
  private static final int THREAD_TYPE_LISTENER = 2;
  private static final int THREAD_TYPE_CALLER = 3;

  private FaasIxServer mainThread;
  private int threadType;
  private String[] packageNames;
  private ServerSocket serverSocket;
  private Socket callSocket;
  private HashMap<String,Class> classNameMap;


  private class FaasIxException extends RuntimeException {
    FaasIxException(String message) {
      super(message);
    }
  }

  FaasIxServer(int threadType, FaasIxServer mainThread) {
    this.threadType = threadType;
    this.mainThread = mainThread;
    if (mainThread == null && threadType == THREAD_TYPE_MAIN) {
      this.mainThread = this;
    }
  }

  public static void main(String args[]) {
    FaasIxServer mainThread = new FaasIxServer(THREAD_TYPE_MAIN, null);
    mainThread.packageNames = args;
    mainThread.start();
    try {
      mainThread.join();
    }
    catch (InterruptedException e) {
      throw new RuntimeException(e);
    }
  }

  private ServerSocket getServerSocket() {
    return this.mainThread.serverSocket;
  }

  public void run() {
    System.out.println("RUNNING THREAD: "+String.valueOf(this.threadType));
    try {
      if (this.threadType == THREAD_TYPE_MAIN) {
        runServer();
      }
      else if (this.threadType == THREAD_TYPE_RELAY) {
        int relayExit = runRelayThread();
        if (relayExit != 0) {
          throw new RuntimeException("Relay exited with status "+String.valueOf(relayExit));
        }
      }
      else if (this.threadType == THREAD_TYPE_LISTENER) {
        setupRelay();
        listenForCalls();
      }
      else if (this.threadType == THREAD_TYPE_CALLER) {
        runCall();
      }
      else {
        throw new RuntimeException("Invalid thread type: "+String.valueOf(this.threadType));
      }
    }
    catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  private void runServer() throws InterruptedException, IOException {
    System.out.println("running");
    this.mainThread.serverSocket = new ServerSocket(0);

    FaasIxServer relayThread = new FaasIxServer(THREAD_TYPE_RELAY, this.mainThread);
    relayThread.start();

    FaasIxServer listenerThread = new FaasIxServer(THREAD_TYPE_LISTENER, this.mainThread);
    listenerThread.packageNames = this.packageNames;
    listenerThread.start();

    relayThread.join();
    getServerSocket().close();
  }

  private int runRelayThread() throws InterruptedException, IOException {
    InetSocketAddress addr = (InetSocketAddress)getServerSocket().getLocalSocketAddress();
    String[] childCmd = new String[4];
    // childCmd[0] = "./dev-faas-ix-polyglot-relay";
    // childCmd[1] = "tcp";
    // childCmd[2] = addr.getAddress().getHostAddress();
    // childCmd[3] = String.valueOf(addr.getPort());
    childCmd[0] = "./faas-ix-polyglot-relay";
    childCmd[1] = "tcp:"+addr.getAddress().getHostAddress()+":"+String.valueOf(addr.getPort());
    ProcessBuilder pb = new ProcessBuilder(childCmd);
    pb.redirectOutput(ProcessBuilder.Redirect.INHERIT).redirectError(ProcessBuilder.Redirect.INHERIT);
    Process relayProcess = pb.start();
    int exit = relayProcess.waitFor();
    return exit;
  }

  private void setupRelay() throws IOException {
    Socket socket = getServerSocket().accept();
    System.out.println("accepted");
    JsonReader jr = Json.createReader(socket.getInputStream());
    JsonObject config = jr.readObject();
    // System.out.println(config.toString());
    JsonWriter jw = Json.createWriter(socket.getOutputStream());
    JsonObject exports = loadExports();
    System.out.println(exports.toString());
    jw.writeObject(exports);
    jr.close();
    jw.close();
    socket.close();
  }

  private JsonObject loadExports() {
    HashMap<String,HashMap> exports = getExports();

    // JsonObjectBuilder builder = Json.createObjectBuilder();
    // for (Class klass : classList) {
    //   JsonObjectBuilder methodBuilder = Json.createObjectBuilder();
    //   for (Method method : klass.getMethods()) {
    //     if (true) {
    //       methodBuilder.add(method.getName(), methodVal);
    //     }
    //   }
    //   classVal = Json.createObjectBuilder()
    //     .add("exports", Json.createObjectBuilder()
    //          .add("isFunc", false)
    //          .add("atts", methodBuilder))
    //     .build();
    //   className = klass.getName();
    //   builder.add(className, classVal);
    //   classNames.put(className, null);
    //   // Alias without package prefix.
    //   className = className.substring(className.lastIndexOf('.')+1);
    //   // Alias inner classes to replace '$' with '.'.
    // }

    JsonObjectBuilder builder = Json.createObjectBuilder();
    for (Map.Entry<String,HashMap> tld_info : exports.entrySet()) {
      builder.add(tld_info.getKey(), Json.createObjectBuilder()
                  .add("exports", Json.createObjectBuilder()
                       .add("isFunc", false)
                       .add("atts", metaForAtts(tld_info.getValue()))));
    }
    return builder.build();
  }

  private HashMap<String,HashMap> getExports() {
    ArrayList<Class> classList = new ArrayList<Class>();
    for (String pkg : this.packageNames) {
      try {
        classList.addAll(getClassesForPackage(pkg));
      }
      catch (Exception e) {
        e.printStackTrace();
      }
    }
    return exportsFromClasses(classList);
  }

  private JsonObject metaForAtts(HashMap<String,Object> atts) {
    JsonObject methodVal = Json.createObjectBuilder().add("isFunc", true).build();
    JsonObjectBuilder builder = Json.createObjectBuilder();
    String key;
    Object val;
    for (Map.Entry<String,Object> entry : atts.entrySet()) {
      key = entry.getKey();
      val = entry.getValue();
      if (val instanceof Method) {
        builder.add(key, methodVal);
      }
      else {
        builder.add(key, Json.createObjectBuilder()
                    .add("isFunc", false)
                    .add("atts", metaForAtts((HashMap)val)));
      }
    }
    return builder.build();
  }

  private HashMap<String,HashMap> exportsFromClasses(ArrayList<Class> classList) {
    // hash: { key => value }
    // key: (String) token of package or class name, or method name.
    // value: {
    //    (String)methodName => (Method)
    //    (String)token => (HashMap)
    // }
    HashMap<String,HashMap> exports = new HashMap<String,HashMap>();
    HashMap methodMap;
    int modifiers;
    String className;
    for (Class klass : classList) {
      methodMap = new HashMap<String,Object>();
      for (Method method : klass.getMethods()) {
        modifiers = method.getModifiers();
        if (Modifier.isStatic(modifiers) && Modifier.isPublic(modifiers)) {
          methodMap.put(method.getName(), method);
        }
      }
      setClassExports(klass, exports, methodMap);
    }
    return exports;
  }

  private void setClassExports(Class klass, HashMap<String,HashMap> exports, HashMap methodMap) {
    String className = klass.getName();
    addMethodsToExports(klass, className, exports, methodMap);
    // Alias without package prefix.
    className = className.substring(className.lastIndexOf('.')+1);
    addMethodsToExports(klass, className, exports, methodMap);
  }

  private void addMethodsToExports(Class klass, String className, HashMap<String,HashMap> exports, HashMap methodMap) {
    HashMap subMap = exports;
    String clsName = className.replace('$','.'); // Inner classes.
    setClassNameMap(clsName, klass);
    for (String name : clsName.split("\\.")) {
      if (!subMap.containsKey(name)) {
        subMap.put(name, new HashMap<String,HashMap>());
      }
      subMap = (HashMap)subMap.get(name);
    }
    subMap.putAll(methodMap);
  }

  private void setClassNameMap(String name, Class klass) {
    getClassNameMap().put(name, klass);
  }

  private HashMap<String,Class> getClassNameMap() {
    if (this.mainThread.classNameMap == null) {
      this.mainThread.classNameMap = new HashMap<String,Class>();
    }
    return this.mainThread.classNameMap;
  }


  private void listenForCalls() throws IOException {
    while (true) {
      Socket socket = getServerSocket().accept();
      FaasIxServer callerThread = new FaasIxServer(THREAD_TYPE_CALLER, this.mainThread);
      callerThread.callSocket = socket;
      callerThread.start();
    }
  }

  private void runCall() throws IOException {
    JsonReader jr = Json.createReader(this.callSocket.getInputStream());
    Exception err = null;
    JsonValue data = null;
    try {
      JsonObject callData = jr.readObject();
      String api = callData.getString("api");
      JsonArray args = callData.getJsonArray("args");
      data = toJsonValue(callMethod(api, args));
    }
    catch (Exception e) {
      e.printStackTrace();
      err = e;
    }
    JsonObjectBuilder respBuilder = Json.createObjectBuilder();
    if (err != null) { respBuilder.add("err", err.toString()); }
    if (data != null) { respBuilder.add("data", data); }
    JsonObject resp = respBuilder.build();
    System.out.println(resp.toString());
    JsonWriter jw = Json.createWriter(this.callSocket.getOutputStream());
    jw.writeObject(resp);
    jr.close();
    jw.close();
    this.callSocket.close();
  }

  private Object callMethod(String api, JsonArray args) throws ClassNotFoundException, IllegalAccessException, InvocationTargetException {
    System.out.println("api: "+api.toString()+", args: "+args.toString());
    Object[] invokeArgs = fromJsonArray(args);
    Method method = getMethod(api, invokeArgs);
    return method.invoke(null, invokeArgs);
  }

  private Method getMethod(String api, Object[] args) throws ClassNotFoundException {
    int dot = api.lastIndexOf('.');
    String className = api.substring(0,dot);
    String methodName = api.substring(dot+1);
    // return Class.forName(className).getMethod(methodName, classForArgs(args));
    Method method = null;
    for (Method meth : getClassNameMap().get(className).getMethods()) {
      if (methodName.equals(meth.getName())) {
        if (method == null || meth.getParameterTypes().length == args.length) {
          // TODO: Check parameter types.
          method = meth;
        }
      }
    }
    return method;
  }

  private JsonValue toJsonValue(Object o) {
    JsonValue jv;
    JsonArrayBuilder builder = Json.createArrayBuilder();
    if (o == null) {
      jv = JsonValue.NULL;
    }
    else if (o instanceof String) {
      jv = builder.add((String)o).build().getJsonString(0);
    }
    else if (o instanceof Boolean) {
      jv = builder.add((Boolean)o).build().getBoolean(0) ? JsonValue.TRUE : JsonValue.FALSE;
    }
    else if (o instanceof Number) {
      jv = builder.add(((Number)o).doubleValue()).build().getJsonNumber(0);
    }
    else if (o instanceof Collection || o.getClass().isArray()) {
      jv = toJsonArray(builder, o);
    }
    else if (o instanceof Map) {
      // http://stackoverflow.com/questions/509076/how-do-i-address-unchecked-cast-warnings
      @SuppressWarnings("unchecked")
      Map<String,Object> map = (Map<String,Object>)o;
      jv = toJsonObject(map);
    }
    else {
      throw new FaasIxException("Unable to convert Object to JsonValue: "+o.getClass().toString());
    }
    return jv;
  }

  private JsonArray toJsonArray(JsonArrayBuilder builder, Object o) {
    Object[] arr;
    if (o instanceof Collection) {
      arr = ((Collection)o).toArray();
    }
    else {
      arr = (Object[])o;
    }
    for (Object a : arr) {
      builder.add(toJsonValue(a));
    }
    return builder.build();
  }

  private JsonObject toJsonObject(Map<String,Object> map) {
    JsonObjectBuilder builder = Json.createObjectBuilder();
    for (Map.Entry entry: map.entrySet()) {
      builder.add(entry.getKey().toString(), toJsonValue(entry.getValue()));
    }
    return builder.build();
  }

  private Object[] fromJsonArray(JsonArray ja) {
    Object[] args = ja.toArray();
    for (int i = 0 ; i < args.length ; i++) {
      args[i] = fromJsonValue((JsonValue)args[i]);
    }
    return args;
  }

  private HashMap fromJsonObject(JsonObject jo) {
    HashMap<String,Object> hashMap = new HashMap<String,Object>(jo);
    for (Map.Entry<String,Object> entry: hashMap.entrySet()) {
      entry.setValue(fromJsonValue((JsonValue)entry.getValue()));
    }
    return hashMap;
  }

  private Object fromJsonValue(JsonValue jv) {
    Object o;
    if (jv instanceof JsonArray) {
      o = fromJsonArray((JsonArray)jv);
    }
    else if (jv instanceof JsonNumber) {
      o = ((JsonNumber)jv).bigDecimalValue();
    }
    else if (jv instanceof JsonObject) {
      o = fromJsonObject((JsonObject)jv);
    }
    else if (jv instanceof JsonString) {
      o = ((JsonString)jv).getString();
    }
    else if (jv == JsonValue.TRUE) {
      o = Boolean.TRUE;
    }
    else if (jv == JsonValue.FALSE) {
      o = Boolean.FALSE;
    }
    else if (jv == JsonValue.NULL) {
      o = null;
    }
    else {
      throw new FaasIxException("Unable to convert JsonValue to Object: "+jv.getClass().toString());
    }
    return o;
  }



  /////////////////////////////////////
  /////////////////////////////////////
  // http://stackoverflow.com/questions/520328/can-you-find-all-classes-in-a-package-using-reflection/22462785#22462785
  /////////////////////////////////////
  /////////////////////////////////////

  /**
   * Private helper method
   * 
   * @param directory
   *            The directory to start with
   * @param pckgname
   *            The package name to search for. Will be needed for getting the
   *            Class object.
   * @param classes
   *            if a file isn't loaded but still is in the directory
   * @throws ClassNotFoundException
   */
  private static void checkDirectory(File directory, String pckgname,
                                     ArrayList<Class<?>> classes) throws ClassNotFoundException {
    File tmpDirectory;

    if (directory.exists() && directory.isDirectory()) {
      final String[] files = directory.list();

      for (final String file : files) {
        if (file.endsWith(".class")) {
          try {
            classes.add(Class.forName(pckgname + '.'
                                      + file.substring(0, file.length() - 6)));
          } catch (final NoClassDefFoundError e) {
            // do nothing. this class hasn't been found by the
            // loader, and we don't care.
          }
        } else if ((tmpDirectory = new File(directory, file))
                   .isDirectory()) {
          checkDirectory(tmpDirectory, pckgname + "." + file, classes);
        }
      }
    }
  }

  /**
   * Private helper method.
   * 
   * @param connection
   *            the connection to the jar
   * @param pckgname
   *            the package name to search for
   * @param classes
   *            the current ArrayList of all classes. This method will simply
   *            add new classes.
   * @throws ClassNotFoundException
   *             if a file isn't loaded but still is in the jar file
   * @throws IOException
   *             if it can't correctly read from the jar file.
   */
  private static void checkJarFile(JarURLConnection connection,
                                   String pckgname, ArrayList<Class<?>> classes)
    throws ClassNotFoundException, IOException {
    final JarFile jarFile = connection.getJarFile();
    final Enumeration<JarEntry> entries = jarFile.entries();
    String name;

    for (JarEntry jarEntry = null; entries.hasMoreElements()
           && ((jarEntry = entries.nextElement()) != null);) {
      name = jarEntry.getName();

      if (name.contains(".class")) {
        name = name.substring(0, name.length() - 6).replace('/', '.');

        if (name.contains(pckgname)) {
          classes.add(Class.forName(name));
        }
      }
    }
  }

  /**
   * Attempts to list all the classes in the specified package as determined
   * by the context class loader
   * 
   * @param pckgname
   *            the package name to search
   * @return a list of classes that exist within that package
   * @throws ClassNotFoundException
   *             if something went wrong
   */
  public static ArrayList<Class<?>> getClassesForPackage(String pckgname)
    throws ClassNotFoundException {
    final ArrayList<Class<?>> classes = new ArrayList<Class<?>>();

    try {
      final ClassLoader cld = Thread.currentThread()
        .getContextClassLoader();

      if (cld == null)
        throw new ClassNotFoundException("Can't get class loader.");

      final Enumeration<URL> resources = cld.getResources(pckgname
                                                          .replace('.', '/'));
      URLConnection connection;

      for (URL url = null; resources.hasMoreElements()
             && ((url = resources.nextElement()) != null);) {
        try {
          connection = url.openConnection();

          if (connection instanceof JarURLConnection) {
            checkJarFile((JarURLConnection) connection, pckgname,
                         classes);
          }
          else if (url.getProtocol().equals("file")) {
            try {
              checkDirectory(
                             new File(URLDecoder.decode(url.getPath(),
                                                        "UTF-8")), pckgname, classes);
            } catch (final UnsupportedEncodingException ex) {
              throw new ClassNotFoundException(
                                pckgname
                                + " does not appear to be a valid package (Unsupported encoding)",
                                ex);
            }
          } else {
            throw new ClassNotFoundException(pckgname + " ("
                                             + url.toString()
                                             + ") does not appear to be a valid package");
          }
        } catch (final IOException ioex) {
          throw new ClassNotFoundException(
                        "IOException was thrown when trying to get all resources for "
                        + pckgname, ioex);
        }
      }
    } catch (final NullPointerException ex) {
      throw new ClassNotFoundException(
                pckgname
                + " does not appear to be a valid package (Null pointer exception)",
                ex);
    } catch (final IOException ioex) {
      throw new ClassNotFoundException(
                "IOException was thrown when trying to get all resources for "
                + pckgname, ioex);
    }

    return classes;
  }



}
