var Base = require('../base')
  , Crypto = require('./crypto')
  , Publisher = require('./publisher')
  , PolyglotRelay = require('./polyglot_relay')
var io = require('socket.io-client')

var EVENTS = [ 'connect', 'connect_error', 'connect_timeout', 'reconnect', 'reconnect_error', 'reconnect_failed', 'error', 'disconnect', 'open', 'close', 'reconnect_attempt' ]

var debug = Base.logger('ix-connector')

var socket, proxyData
var isRelay = false

exports.connectToProxy = function(data, cb_err_data){
  debug('connectToProxy', data)
  proxyData = data
  console.log('Connecting to '+data.proxyUrl+'...')
  socket = io(data.proxyUrl)
  setDebugHandlers(socket, EVENTS)
  socket.on('connect', function(){ onConnect(cb_err_data) })
  socket.on('ix-call', onIxCall)
  socket.on('ix-ping', onIxPing)
  socket.on('ix-x', onIxX)
}

exports.connectRelayToProxy = function(data, cb_err_data){
  debug('connectRelayToProxy', data)
  isRelay = true
  exports.connectToProxy(data, cb_err_data)
}

var setDebugHandlers = function(socket, events){
  var i, event
  for (var i = 0; i < events.length; i++) {
    event = events[i]
    socket.on(event, debugHandlerFor(event))
    socket.io.on(event, debugHandlerFor('manager:'+event)) // Manager events
  }
}

var debugHandlerFor = function(type){
  return function(){
    var args = Array.prototype.slice.call(arguments, 0)
    args.unshift('event:'+type)
    debug.apply(this, args)
  }
}

var onConnect = function(cb_err_data){
  debug('connect', arguments)
  socket.emit('set-client', { apiConfig: Base.apiConfig(), proxyUrl: proxyData.proxyUrl, proxyExport: proxyData.proxyExport }, function(err, data){
    if (err) {
      console.log('Connect error: Unable to complete connection to '+proxyData.proxyUrl+': '+JSON.stringify(err))
    }
    else {
      console.log('Connected to '+proxyData.proxyUrl+'.')
    }
    // TODO: if (err) { retryConnection }
    debug('connect result', err, data)
    cb_err_data(err, data)
  })
}

var onIxCall = function(data, responder){
  debug('ix-call', arguments)
  responder = debug.wrap(responder)
  try {
    if (isRelay) {
      Publisher.ixCall(data, responder, PolyglotRelay.invoker)
    }
    else {
      Publisher.ixCallNodejs(data, responder)
    }
  }
  catch(e) {
    responder(e)
  }
}

var onIxPing = function(replyData, responder){
  debug('ix-ping', replyData)
  responder(null, replyData)
}

var onIxX = function(cipherParams, responder){
  debug('ix-x', cipherParams)
  Crypto.decrypt(cipherParams, function(err, data){
    if (err) { responder.call(err); return }
    onIxCall(data, Crypto.encryptingResponder(responder))
  })
}
